@@ -43,7 +43,7 @@ class ParsingSexException(Exception):
 
 
 class AgeException(Exception):
-    """Exception for invalid sex"""
+    """Exception for invalid age"""
 
 
 class PhenotypeNameException(Exception):
@@ -51,7 +51,17 @@ class PhenotypeNameException(Exception):
 
 
 class RelationshipException(Exception):
-    """Exception for a relationship between non existing phenotypes"""
+    """Exception for a relationship between non existing phenotypes or generic errors in creating a relationship"""
+
+    def __init__(self, error_message, r=None):
+        self.message = error_message
+        if r is not None:
+            self.message += f". Code: {r.status_code}, response: {get_response(r)}"
+        super().__init__(self.message)
+
+
+class GeodataException(Exception):
+    """Exception for errors in geodata"""
 
 
 class TechnicalMalformedException(Exception):
@@ -66,6 +76,66 @@ class TechnicalAssociationException(Exception):
     """Exception for technicals with a non existing dataset associated"""
 
 
+class ResourceCreationException(Exception):
+    """Exception for errors in creating resources"""
+
+    def __init__(self, error_message, r=None):
+        self.message = error_message
+        if r is not None:
+            self.message += f". Code: {r.status_code}, response: {get_response(r)}"
+        super().__init__(self.message)
+
+
+class ResourceRetrievingException(Exception):
+    """Exception for errors in retrieving resources"""
+
+    def __init__(self, error_message, r=None):
+        self.message = error_message
+        if r is not None:
+            self.message += f". Code: {r.status_code}, response: {get_response(r)}"
+        super().__init__(self.message)
+
+
+class ResourceAssignationException(Exception):
+    """Exception for errors in assignating resources to other resources (ex. phenotypes to a dataset)"""
+
+    def __init__(self, error_message, r=None):
+        self.message = error_message
+        if r is not None:
+            self.message += f". Code: {r.status_code}, response: {get_response(r)}"
+        super().__init__(self.message)
+
+
+class ResourceModificationException(Exception):
+    """Exception for errors in modify resources"""
+
+    def __init__(self, error_message, r=None):
+        self.message = error_message
+        if r is not None:
+            self.message += f". Code: {r.status_code}, response: {get_response(r)}"
+        super().__init__(self.message)
+
+
+class UploadInitException(Exception):
+    """Exception for errors in initializing an upload"""
+
+    def __init__(self, error_message, r=None):
+        self.message = error_message
+        if r is not None:
+            self.message += f". Code: {r.status_code}, response: {get_response(r)}"
+        super().__init__(self.message)
+
+
+class UploadException(Exception):
+    """Exception for errors in uploading a file"""
+
+    def __init__(self, error_message, r=None):
+        self.message = error_message
+        if r is not None:
+            self.message += f". Code: {r.status_code}, response: {get_response(r)}"
+        super().__init__(self.message)
+
+
 @contextmanager
 def pfx_to_pem(pfx_path: Path, pfx_password: str) -> Generator[str, None, None]:
     """Decrypts the .pfx file to be used with requests."""
@@ -149,6 +219,11 @@ def error(text: str, r: Optional[requests.Response] = None) -> None:
     return None
 
 
+def warning(text: str) -> None:
+    typer.secho(text, fg=typer.colors.YELLOW)
+    return None
+
+
 def success(text: str) -> None:
     typer.secho(text, fg=typer.colors.GREEN)
     return None
@@ -390,7 +465,7 @@ def parse_file_tech(
 
 def version_callback(value: bool) -> None:
     if value:
-        typer.echo("NIG Upload version: 0.3")
+        typer.echo("NIG Upload version: 0.4")
         raise typer.Exit()
 
 
@@ -458,9 +533,391 @@ def get_ip() -> str:
     return urllib.request.urlopen("https://ident.me").read().decode("utf8")
 
 
+def validate_study(study: Path) -> Optional[Dict[str, Any]]:
+
+    study_tree: Dict[str, Any] = {
+        "name": study.name,
+        "phenotypes": "",
+        "technicals": "",
+        "datasets": {},
+    }
+
+    for d in study.iterdir():
+        if d.is_dir():
+            for dat in d.iterdir():
+                if (
+                    dat.is_file()
+                    and dat.name.endswith(".fastq.gz")
+                    and dat.stat().st_size >= 1
+                ):
+                    study_tree["datasets"].setdefault(d.name, [])
+                    study_tree["datasets"][d.name].append(dat)
+            if (
+                study_tree["datasets"].get(d.name)
+                and len(study_tree["datasets"][d.name]) > 2
+            ):
+                # the dataset is invalid because contains too many fastq
+                warning(
+                    f"Upload of {study.name} skipped: Dataset {d.name} contains too many fastq files: max allowed files are 2 per dataset"
+                )
+                return None
+
+    if not study_tree["datasets"]:
+        warning(
+            f"Upload of {study.name} skipped: No files found for upload in: {study}"
+        )
+        return None
+
+    pedigree = study.joinpath("pedigree.txt")
+    if pedigree.is_file():
+        try:
+            phenotypes_list, relationships = parse_file_ped(
+                pedigree, study_tree["datasets"]
+            )
+        except (
+            PhenotypeMalformedException,
+            PhenotypeNameException,
+            HPOException,
+            ParsingSexException,
+            AgeException,
+            RelationshipException,
+        ) as exc:
+            warning(f"Upload of {study.name} skipped: {exc}")
+            return None
+
+        study_tree["phenotypes"] = phenotypes_list
+        study_tree["relationships"] = relationships
+
+    technical = study.joinpath("technical.txt")
+    if technical.is_file():
+        try:
+            technicals_list = parse_file_tech(technical, study_tree["datasets"])
+        except (
+            TechnicalMalformedException,
+            UnknownPlatformException,
+            TechnicalAssociationException,
+        ) as exc:
+            warning(f"Upload of {study.name} skipped: {exc}")
+            return None
+
+        study_tree["technicals"] = technicals_list
+
+    return study_tree
+
+
+def get_technical_uuid(
+    study_tree: Dict[str, Any], dataset_name: str, technicals_uuid: Dict[str, str]
+) -> Optional[str]:
+    tech_uuid: Optional[str] = None
+    if len(study_tree["technicals"]) > 1:
+        for tech in study_tree["technicals"]:
+            if dataset_name in tech["datasets"]:
+                tech_uuid = technicals_uuid[tech["properties"]["name"]]
+                break
+    else:
+        if (
+            "datasets" not in study_tree["technicals"][0].keys()
+            or "datasets" in study_tree["technicals"][0].keys()
+            and dataset_name in study_tree["technicals"][0]["datasets"]
+        ):
+            tech_uuid = technicals_uuid[
+                study_tree["technicals"][0]["properties"]["name"]
+            ]
+    return tech_uuid
+
+
+def upload_study(
+    study_tree: Dict[str, Any],
+    url: str,
+    headers: Dict[str, str],
+    certfile: Path,
+    certpwd: str,
+    chunk_size: int,
+    IP_ADDR: str,
+) -> None:
+    study_name = study_tree["name"]
+    r = request(
+        method=POST,
+        url=f"{url}api/study",
+        headers=headers,
+        certfile=certfile,
+        certpwd=certpwd,
+        data={"name": study_name, "description": ""},
+    )
+    if r.status_code != 200:
+        raise ResourceCreationException("Study creation failed", r)
+
+    success(f"Succesfully created study {study_name}")
+
+    study_uuid = r.json()
+
+    # create phenotypes
+    if study_tree["phenotypes"]:
+        # get geodata list
+        r = request(
+            method=POST,
+            url=f"{url}api/study/{study_uuid}/phenotypes",
+            headers=headers,
+            certfile=certfile,
+            certpwd=certpwd,
+            data={"get_schema": True},
+        )
+        if r.status_code != 200:
+            raise ResourceRetrievingException("Can't retrieve geodata list", r)
+
+        for el in r.json():
+            if el["key"] == "birth_place":
+                geodata = el["options"]
+                break
+        phenotypes_uuid: Dict[str, str] = {}
+        for phenotype in study_tree["phenotypes"]:
+            # get the birth_place
+            if phenotype.get("birth_place_name"):
+                for geo_id, name in geodata.items():
+                    if name == phenotype["birth_place_name"]:
+                        phenotype["birth_place"] = geo_id
+                        break
+                if "birth_place" not in phenotype.keys():
+                    raise GeodataException(
+                        f"Error for phenotype {phenotype['name']}: {phenotype['birth_place_name']} birth place not found"
+                    )
+
+                # delete birth_place_name key
+                del phenotype["birth_place_name"]
+
+            r = request(
+                method=POST,
+                url=f"{url}api/study/{study_uuid}/phenotypes",
+                headers=headers,
+                certfile=certfile,
+                certpwd=certpwd,
+                data=phenotype,
+            )
+            if r.status_code != 200:
+                raise ResourceCreationException("Phenotype creation failed", r)
+
+            success(f"Succesfully created phenotype {phenotype['name']}")
+
+            # add the uuid in the phenotype uuid dictionary
+            phenotypes_uuid[phenotype["name"]] = r.json()
+
+    # create phenotypes relationships
+    if "relationships" in study_tree.keys():
+        for son, parent_list in study_tree["relationships"].items():
+            son_uuid = phenotypes_uuid.get(son)
+            for parent in parent_list:
+                parent_uuid = phenotypes_uuid.get(parent)
+                r = request(
+                    method=POST,
+                    url=f"{url}api/phenotype/{son_uuid}/relationships/{parent_uuid}",
+                    headers=headers,
+                    certfile=certfile,
+                    certpwd=certpwd,
+                    data={},
+                )
+                if r.status_code != 200:
+                    raise RelationshipException("Phenotype relationship failed", r)
+
+                success(f"Succesfully created relationship between {son} and {parent}")
+
+    # create technicals
+    technicals_uuid: Dict[str, str] = {}
+    if study_tree["technicals"]:
+        for technical in study_tree["technicals"]:
+            r = request(
+                method=POST,
+                url=f"{url}api/study/{study_uuid}/technicals",
+                headers=headers,
+                certfile=certfile,
+                certpwd=certpwd,
+                data=technical["properties"],
+            )
+            if r.status_code != 200:
+                raise ResourceCreationException("Technical creation failed", r)
+
+            success(f"Succesfully created technical {technical['properties']['name']}")
+
+            # add the uuid in the technical uuid dictionary
+            technicals_uuid[technical["properties"]["name"]] = r.json()
+
+    for dataset_name, files in study_tree["datasets"].items():
+        r = request(
+            method=POST,
+            url=f"{url}api/study/{study_uuid}/datasets",
+            headers=headers,
+            certfile=certfile,
+            certpwd=certpwd,
+            data={"name": dataset_name, "description": ""},
+        )
+
+        if r.status_code != 200:
+            raise ResourceCreationException("Dataset creation failed", r)
+
+        success(f"Succesfully created dataset {dataset_name}")
+        uuid = r.json()
+
+        #  connect the phenotype to the dataset
+        if dataset_name in phenotypes_uuid.keys():
+            phen_uuid = phenotypes_uuid[dataset_name]
+            r = request(
+                method=PUT,
+                url=f"{url}api/dataset/{uuid}",
+                headers=headers,
+                certfile=certfile,
+                certpwd=certpwd,
+                data={"phenotype": phen_uuid},
+            )
+            if r.status_code != 204:
+                raise ResourceAssignationException(
+                    "Can't assign a phenotype to the dataset", r
+                )
+
+            success(f"Succesfully assigned phenotype to dataset {dataset_name}")
+
+        #  connect the technical to the dataset
+        if study_tree["technicals"]:
+            tech_uuid = get_technical_uuid(study_tree, dataset_name, technicals_uuid)
+
+            if tech_uuid:
+                r = request(
+                    method=PUT,
+                    url=f"{url}api/dataset/{uuid}",
+                    headers=headers,
+                    certfile=certfile,
+                    certpwd=certpwd,
+                    data={"technical": tech_uuid},
+                )
+                if r.status_code != 204:
+                    raise ResourceAssignationException(
+                        "Can't assign a technical to the dataset", r
+                    )
+
+                success(f"Succesfully assigned technical to dataset {dataset_name}")
+
+        for file in files:
+            # get the data for the upload request
+            filename = file.name
+            filesize = file.stat().st_size
+            mimeType = MimeTypes().guess_type(str(file))
+            lastModified = int(file.stat().st_mtime)
+
+            data = {
+                "name": filename,
+                "mimeType": mimeType,
+                "size": filesize,
+                "lastModified": lastModified,
+            }
+
+            # init the upload
+            r = request(
+                method=POST,
+                url=f"{url}api/dataset/{uuid}/files/upload",
+                headers=headers,
+                certfile=certfile,
+                certpwd=certpwd,
+                data=data,
+            )
+
+            if r.status_code != 201:
+                raise UploadInitException("Can't start the upload", r)
+
+            success("Upload succesfully initialized")
+
+            chunk = chunk_size * 1024 * 1024
+            range_start = -1
+            prev_position = 0
+
+            with open(file, "rb") as f:
+                start = datetime.now()
+                with typer.progressbar(length=filesize, label="Uploading") as progress:
+                    while True:
+
+                        prev_position = f.tell()
+                        read_data = f.read(chunk)
+                        # No more data read from the file
+                        if not read_data:
+                            break
+
+                        range_start += 1
+
+                        range_max = min(range_start + chunk, filesize)
+
+                        content_range = f"bytes {range_start}-{range_max}/{filesize}"
+                        headers["Content-Range"] = content_range
+
+                        try:
+
+                            r = request(
+                                method=PUT,
+                                url=f"{url}api/dataset/{uuid}/files/upload/{filename}",
+                                headers=headers,
+                                certfile=certfile,
+                                certpwd=certpwd,
+                                data=read_data,
+                            )
+                        except (
+                            requests.exceptions.ConnectionError,
+                            requests.exceptions.ReadTimeout,
+                        ) as r:
+
+                            IP = get_ip()
+                            if IP != IP_ADDR:
+                                return error(
+                                    f"\nUpload failed due to a network error ({r})"
+                                    f"\nYour IP address changed from {IP_ADDR} to {IP}."
+                                    "\nDue to security policies the upload"
+                                    " can't be retried"
+                                )
+                            else:
+                                error(f"Upload Failed, retrying ({str(r)})")
+                                f.seek(prev_position)
+                                range_start -= 1
+                                continue
+
+                        if r.status_code != 206:
+                            if r.status_code == 200:
+                                # upload is complete
+                                progress.update(filesize)
+                                break
+                            raise UploadException("Upload Failed", r)
+
+                        progress.update(chunk)
+                        # update the range variable
+                        range_start += chunk
+
+                end = datetime.now()
+                seconds = (end - start).seconds or 1
+
+                t = get_time(seconds)
+                s = get_speed(filesize / seconds)
+                if r.status_code != 200:
+                    raise UploadException(f"Upload Failed in {t} ({s})", r)
+
+                success(f"Upload succesfully completed in {t} ({s})")
+
+        # set the status of the dataset as "UPLOAD COMPLETED"
+        r = request(
+            method=PATCH,
+            url=f"{url}api/dataset/{uuid}",
+            headers=headers,
+            certfile=certfile,
+            certpwd=certpwd,
+            data={"status": "UPLOAD COMPLETED"},
+        )
+        if r.status_code != 204:
+            raise ResourceModificationException(
+                "Can't set the status to the dataset", r
+            )
+
+        success(f"Succesfully set UPLOAD COMPLETE to {dataset_name}")
+
+
 @app.command()
 def upload(
-    study: Path = typer.Argument(..., help="Path to the study"),
+    study: Path = typer.Option(None, help="Path to the study"),
+    studies: Path = typer.Option(
+        None, help="Path to the main folder containing the studies directories"
+    ),
     url: str = typer.Option(..., prompt="Server URL", help="Server URL"),
     username: str = typer.Option(..., prompt="Your username"),
     pwd: str = typer.Option(..., prompt="Your password", hide_input=True),
@@ -486,6 +943,11 @@ def upload(
     ),
 ) -> None:
 
+    if not study and not studies:
+        return error(
+            "A path to a study or to a directory of studies has to be specified"
+        )
+
     if not url.startswith("https:"):
         url = f"https://{url}"
     if not url.endswith("/"):
@@ -497,72 +959,6 @@ def upload(
     if chunk_size > 16:
         return error(f"The specified chunk size is too large: {chunk_size}")
 
-    # check if the input file exists
-    if not study.exists():
-        return error(f"The specified study does not exists: {study}")
-
-    study_tree: Dict[str, Any] = {
-        "name": study.name,
-        "phenotypes": "",
-        "technicals": "",
-        "datasets": {},
-    }
-
-    for d in study.iterdir():
-        if d.is_dir():
-            for dat in d.iterdir():
-                if (
-                    dat.is_file()
-                    and dat.name.endswith(".fastq.gz")
-                    and dat.stat().st_size >= 1
-                ):
-                    study_tree["datasets"].setdefault(d.name, [])
-                    study_tree["datasets"][d.name].append(dat)
-            if (
-                study_tree["datasets"].get(d.name)
-                and len(study_tree["datasets"][d.name]) > 2
-            ):
-                # the dataset is invalid because contains too many fastq
-                return error(
-                    f"Dataset {d.name} contains too many fastq files: max allowed files are 2 per dataset"
-                )
-
-    if not study_tree["datasets"]:
-        return error(f"No files found for upload in: {study}")
-
-    pedigree = study.joinpath("pedigree.txt")
-    phenotypes_uuid: Dict[str, str] = {}
-    technicals_uuid: Dict[str, str] = {}
-    if pedigree.is_file():
-        try:
-            phenotypes_list, relationships = parse_file_ped(
-                pedigree, study_tree["datasets"]
-            )
-        except (
-            PhenotypeMalformedException,
-            PhenotypeNameException,
-            HPOException,
-            ParsingSexException,
-            AgeException,
-            RelationshipException,
-        ) as exc:
-            return error(exc)
-
-        study_tree["phenotypes"] = phenotypes_list
-        study_tree["relationships"] = relationships
-
-    technical = study.joinpath("technical.txt")
-    if technical.is_file():
-        try:
-            technicals_list = parse_file_tech(technical, study_tree["datasets"])
-        except (
-            TechnicalMalformedException,
-            UnknownPlatformException,
-            TechnicalAssociationException,
-        ) as exc:
-            return error(exc)
-        study_tree["technicals"] = technicals_list
-
     try:
         IP_ADDR = get_ip()
         success(f"Your IP address is {IP_ADDR}")
@@ -587,294 +983,104 @@ def upload(
         headers = {"Authorization": f"Bearer {token}"}
         success("Succesfully logged in")
 
-        study_name = study_tree["name"]
+    except RequestMethodError as exc:
+        return error(exc)
+
+    # get a list of the path to the studies to upload
+    studies_to_upload: List[Path] = []
+    if study:
+        # check if the input directory exists
+        if not study.exists():
+            return error(f"The specified study does not exists: {study}")
+        studies_to_upload.append(study)
+    else:
+        # check if the input directory exists
+        if not studies.exists():
+            return error(
+                f"The specified directory containing the studies directories does not exists: {studies}"
+            )
+        for d in studies.iterdir():
+            if d.is_dir():
+                studies_to_upload.append(d)
+        if not studies_to_upload:
+            return error(f"No studies found in {studies}")
+
+    try:
+        # get user studies list
+        existing_studies: Dict[str, str] = {}
         r = request(
-            method=POST,
+            method=GET,
             url=f"{url}api/study",
             headers=headers,
             certfile=certfile,
             certpwd=certpwd,
-            data={"name": study_name, "description": ""},
+            data={},
         )
         if r.status_code != 200:
-            return error("Study creation failed", r)
+            raise ResourceRetrievingException("Can't retrieve user's studies list", r)
 
-        success(f"Succesfully created study {study_name}")
-
-        study_uuid = r.json()
-
-        # create phenotypes
-        if study_tree["phenotypes"]:
-            # get geodata list
-            r = request(
-                method=POST,
-                url=f"{url}api/study/{study_uuid}/phenotypes",
-                headers=headers,
-                certfile=certfile,
-                certpwd=certpwd,
-                data={"get_schema": True},
-            )
-            if r.status_code != 200:
-                return error("Can't retrieve geodata list", r)
-            for el in r.json():
-                if el["key"] == "birth_place":
-                    geodata = el["options"]
-                    break
-            for phenotype in study_tree["phenotypes"]:
-                # get the birth_place
-                if phenotype.get("birth_place_name"):
-                    for geo_id, name in geodata.items():
-                        if name == phenotype["birth_place_name"]:
-                            phenotype["birth_place"] = geo_id
-                            break
-                    if "birth_place" not in phenotype.keys():
-                        return error(
-                            f"Error for phenotype {phenotype['name']}: {phenotype['birth_place_name']} birth place not found"
-                        )
-
-                    # delete birth_place_name key
-                    del phenotype["birth_place_name"]
+        res = r.json()
+        if res:
+            for el in res:
+                existing_studies[el["name"]] = el["uuid"]
 
+        for s in studies_to_upload:
+            # check if the study already exists
+            if s.name in existing_studies.keys():
+                # get the list of the datasets in the study to upload
+                datasets_to_upload: List[str] = []
+                for d in s.iterdir():
+                    if d.is_dir():
+                        datasets_to_upload.append(d.name)
+                # get the list of the datasets of the existing study
+                existing_datasets: List[str] = []
                 r = request(
-                    method=POST,
-                    url=f"{url}api/study/{study_uuid}/phenotypes",
+                    method=GET,
+                    url=f"{url}api/study/{existing_studies[s.name]}/datasets",
                     headers=headers,
                     certfile=certfile,
                     certpwd=certpwd,
-                    data=phenotype,
+                    data={},
                 )
                 if r.status_code != 200:
-                    return error("Phenotype creation failed", r)
-
-                success(f"Succesfully created phenotype {phenotype['name']}")
-
-                # add the uuid in the phenotype uuid dictionary
-                phenotypes_uuid[phenotype["name"]] = r.json()
-
-        # create phenotypes relationships
-        if "relationships" in study_tree.keys():
-            for son, parent_list in study_tree["relationships"].items():
-                son_uuid = phenotypes_uuid.get(son)
-                for parent in parent_list:
-                    parent_uuid = phenotypes_uuid.get(parent)
-                    r = request(
-                        method=POST,
-                        url=f"{url}api/phenotype/{son_uuid}/relationships/{parent_uuid}",
-                        headers=headers,
-                        certfile=certfile,
-                        certpwd=certpwd,
-                        data={},
-                    )
-                    if r.status_code != 200:
-                        return error("Phenotype relationship failed", r)
-
-                    success(
-                        f"Succesfully created relationship between {son} and {parent}"
+                    raise ResourceRetrievingException(
+                        "Can't retrieve user's datasets list", r
                     )
 
-        # create technicals
-        if study_tree["technicals"]:
-            for technical in study_tree["technicals"]:
-                r = request(
-                    method=POST,
-                    url=f"{url}api/study/{study_uuid}/technicals",
-                    headers=headers,
-                    certfile=certfile,
-                    certpwd=certpwd,
-                    data=technical["properties"],
-                )
-                if r.status_code != 200:
-                    return error("Technical creation failed", r)
-
-                success(
-                    f"Succesfully created technical {technical['properties']['name']}"
-                )
-
-                # add the uuid in the technical uuid dictionary
-                technicals_uuid[technical["properties"]["name"]] = r.json()
-
-        for dataset_name, files in study_tree["datasets"].items():
-            r = request(
-                method=POST,
-                url=f"{url}api/study/{study_uuid}/datasets",
-                headers=headers,
-                certfile=certfile,
-                certpwd=certpwd,
-                data={"name": dataset_name, "description": ""},
-            )
-
-            if r.status_code != 200:
-                return error("Dataset creation failed", r)
-
-            success(f"Succesfully created dataset {dataset_name}")
-            uuid = r.json()
-
-            #  connect the phenotype to the dataset
-            if dataset_name in phenotypes_uuid.keys():
-                phen_uuid = phenotypes_uuid[dataset_name]
-                r = request(
-                    method=PUT,
-                    url=f"{url}api/dataset/{uuid}",
-                    headers=headers,
-                    certfile=certfile,
-                    certpwd=certpwd,
-                    data={"phenotype": phen_uuid},
-                )
-                if r.status_code != 204:
-                    return error("Can't assign a phenotype to the dataset", r)
-
-                success(f"Succesfully assigned phenotype to dataset {dataset_name}")
-
-            #  connect the technical to the dataset
-            if study_tree["technicals"]:
-                tech_uuid: Optional[str] = None
-                if len(study_tree["technicals"]) > 1:
-                    for tech in study_tree["technicals"]:
-                        if dataset_name in tech["datasets"]:
-                            tech_uuid = technicals_uuid[tech["properties"]["name"]]
-                            break
+                res = r.json()
+                if res:
+                    for el in res:
+                        existing_datasets.append(el["name"])
+                # if the two list differs throw an error
+                if not set(datasets_to_upload) == set(existing_datasets):
+                    return error(
+                        f"Study {s.name} already exists but its datasets differ from the already uploaded: Please check"
+                    )
                 else:
-                    if (
-                        "datasets" not in study_tree["technicals"][0].keys()
-                        or "datasets" in study_tree["technicals"][0].keys()
-                        and dataset_name in study_tree["technicals"][0]["datasets"]
-                    ):
-                        tech_uuid = technicals_uuid[
-                            study_tree["technicals"][0]["properties"]["name"]
-                        ]
-                if tech_uuid:
-                    r = request(
-                        method=PUT,
-                        url=f"{url}api/dataset/{uuid}",
-                        headers=headers,
-                        certfile=certfile,
-                        certpwd=certpwd,
-                        data={"technical": tech_uuid},
-                    )
-                    if r.status_code != 204:
-                        return error("Can't assign a technical to the dataset", r)
+                    # the study has already been uploaded
+                    warning(f"Study {s.name} already exists: skipped")
+                    continue
 
-                    success(f"Succesfully assigned technical to dataset {dataset_name}")
-
-            for file in files:
-                # get the data for the upload request
-                filename = file.name
-                filesize = file.stat().st_size
-                mimeType = MimeTypes().guess_type(str(file))
-                lastModified = int(file.stat().st_mtime)
-
-                data = {
-                    "name": filename,
-                    "mimeType": mimeType,
-                    "size": filesize,
-                    "lastModified": lastModified,
-                }
-
-                # init the upload
-                r = request(
-                    method=POST,
-                    url=f"{url}api/dataset/{uuid}/files/upload",
-                    headers=headers,
-                    certfile=certfile,
-                    certpwd=certpwd,
-                    data=data,
-                )
-
-                if r.status_code != 201:
-                    return error("Can't start the upload", r)
-
-                success("Upload succesfully initialized")
-
-                chunk = chunk_size * 1024 * 1024
-                range_start = -1
-                prev_position = 0
-
-                with open(file, "rb") as f:
-                    start = datetime.now()
-                    with typer.progressbar(
-                        length=filesize, label="Uploading"
-                    ) as progress:
-                        while True:
-
-                            prev_position = f.tell()
-                            read_data = f.read(chunk)
-                            # No more data read from the file
-                            if not read_data:
-                                break
-
-                            range_start += 1
-
-                            range_max = min(range_start + chunk, filesize)
-
-                            content_range = (
-                                f"bytes {range_start}-{range_max}/{filesize}"
-                            )
-                            headers["Content-Range"] = content_range
-
-                            try:
-
-                                r = request(
-                                    method=PUT,
-                                    url=f"{url}api/dataset/{uuid}/files/upload/{filename}",
-                                    headers=headers,
-                                    certfile=certfile,
-                                    certpwd=certpwd,
-                                    data=read_data,
-                                )
-                            except (
-                                requests.exceptions.ConnectionError,
-                                requests.exceptions.ReadTimeout,
-                            ) as r:
-
-                                IP = get_ip()
-                                if IP != IP_ADDR:
-                                    return error(
-                                        f"\nUpload failed due to a network error ({r})"
-                                        f"\nYour IP address changed from {IP_ADDR} to {IP}."
-                                        "\nDue to security policies the upload"
-                                        " can't be retried"
-                                    )
-                                else:
-                                    error(f"Upload Failed, retrying ({str(r)})")
-                                    f.seek(prev_position)
-                                    range_start -= 1
-                                    continue
-
-                            if r.status_code != 206:
-                                if r.status_code == 200:
-                                    # upload is complete
-                                    progress.update(filesize)
-                                    break
-                                return error("Upload Failed", r)
-                            progress.update(chunk)
-                            # update the range variable
-                            range_start += chunk
-
-                    end = datetime.now()
-                    seconds = (end - start).seconds or 1
-
-                    t = get_time(seconds)
-                    s = get_speed(filesize / seconds)
-                    if r.status_code != 200:
-                        return error(f"Upload Failed in {t} ({s})", r)
-
-                    success(f"Upload succesfully completed in {t} ({s})")
-
-            # set the status of the dataset as "UPLOAD COMPLETED"
-            r = request(
-                method=PATCH,
-                url=f"{url}api/dataset/{uuid}",
-                headers=headers,
-                certfile=certfile,
-                certpwd=certpwd,
-                data={"status": "UPLOAD COMPLETED"},
+            # validate study
+            study_tree = validate_study(s)
+            if not study_tree:
+                # the study hasn't passed the validation
+                continue
+            # upload the study
+            upload_study(
+                study_tree, url, headers, certfile, certpwd, chunk_size, IP_ADDR
             )
-            if r.status_code != 204:
-                return error("Can't set the status to the dataset", r)
-
-            success(f"Succesfully set UPLOAD COMPLETE to {dataset_name}")
-
-    except RequestMethodError as exc:
+    except (
+        RequestMethodError,
+        ResourceCreationException,
+        ResourceRetrievingException,
+        ResourceAssignationException,
+        ResourceModificationException,
+        UploadInitException,
+        UploadException,
+        GeodataException,
+        RelationshipException,
+    ) as exc:
         return error(exc)
 
     return None
